# FPFS shear estimator
# Copyright 20210905 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

import gc
import os

import fitsio
import galsim
import numpy as np

from .base import setup_custom_logger

_data_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "data",
)


def _galsim_round_sersic(n, sersic_prec):
    return float(int(n / sersic_prec + 0.5)) * sersic_prec


def coord_distort_1(x, y, xref, yref, gamma1, gamma2, kappa=0.0, inverse=False):
    """Distorts coordinates by shear

    Args:
    x (ndarray): input coordinates [x]
    y (ndarray): input coordinates [y]
    xref (float): reference point [x]
    yref (float): reference point [y]
    gamma1 (float): first component of shear distortion
    gamma2 (float): second component of shear distortion
    kappa (float): kappa distortion [default: 0]
    inverse(bool): if true, from source to lens; else, from lens to source

    Returns:
    x2 (ndarray): distorted coordiantes [x]
    y2 (ndarray): distorted coordiantes [y]
    """
    if inverse:
        xu = x - xref
        yu = y - yref
        x2 = (1 - kappa - gamma1) * xu - gamma2 * yu + xref
        y2 = -gamma2 * xu + (1 - kappa + gamma1) * yu + yref
    else:
        u_mag = 1.0 / (1 - kappa) ** 2.0 - gamma1**2.0 - gamma2**2.0
        xu = x - xref
        yu = y - yref
        x2 = ((1 - kappa + gamma1) * xu + gamma2 * yu + xref) * u_mag
        y2 = (gamma2 * xu + (1 - kappa - gamma1) * yu + yref) * u_mag
    return x2, y2


def coord_rotate(x, y, xref, yref, theta):
    """Rotates coordinates by an angle theta (anticlockwise)

    Args:
    x (ndarray): input coordinates [x]
    y (ndarray): input coordinates [y]
    xref (float): reference point [x]
    yref (float): reference point [y]
    theta (float): rotation angle [rads]

    Returns:
    x2 (ndarray): rotated coordiantes [x]
    y2 (ndarray): rotated coordiantes [y]
    """
    xu = x - xref
    yu = y - yref
    x2 = np.cos(theta) * xu - np.sin(theta) * yu + xref
    y2 = np.sin(theta) * xu + np.cos(theta) * yu + yref
    return x2, y2


def generate_cosmos_gal_simple(record, gsparams=None, random_shape=False):
    """Generates COSMOS galaxies; modified version of
    https://github.com/GalSim-developers/GalSim/blob/releases/2.3/galsim/scene.py#L626

    Args:
    record (ndarray): one row of the COSMOS galaxy catalog
    gsparams: an GSParams argument.

    Returns:
    gal: Galsim galaxy
    """

    bdparams = record["bulgefit"]
    sparams = record["sersicfit"]
    use_bulgefit = record["use_bulgefit"]
    if use_bulgefit:
        # Bulge parameters:
        # Minor-to-major axis ratio:
        bulge_hlr = record["hlr"][1]
        bulge_flux = record["flux"][1]
        disk_hlr = record["hlr"][2]
        disk_flux = record["flux"][2]
        bulge = galsim.Gaussian(
            flux=bulge_flux, half_light_radius=bulge_hlr, gsparams=gsparams
        )
        disk = galsim.Exponential(
            flux=disk_flux, half_light_radius=disk_hlr, gsparams=gsparams
        )
        if random_shape:
            # Apply shears for intrinsic shape.
            bulge_q = bdparams[11]
            bulge_beta = bdparams[15] * galsim.radians
            if bulge_q < 1.0:  # pragma: no branch
                bulge = bulge.shear(q=bulge_q, beta=bulge_beta)
            #
            disk_q = bdparams[3]
            disk_beta = bdparams[7] * galsim.radians
            if disk_q < 1.0:  # pragma: no branch
                disk = disk.shear(q=disk_q, beta=disk_beta)
        # Then combine the two components of the galaxy.
        # No center offset is included
        gal = bulge + disk
    else:
        gal_hlr = record["hlr"][0]
        gal_flux = record["flux"][0]

        gal = galsim.Gaussian(
            flux=gal_flux,
            half_light_radius=gal_hlr,
        )
        if random_shape:
            # Apply shears for intrinsic shape.
            gal_q = sparams[3]
            gal_beta = sparams[7] * galsim.radians
            if gal_q < 1.0:
                gal = gal.shear(q=gal_q, beta=gal_beta)
    return gal


def generate_cosmos_gal(record, trunc_ratio=5.0, gsparams=None):
    """Generates COSMOS galaxies; modified version of
    https://github.com/GalSim-developers/GalSim/blob/releases/2.3/galsim/scene.py#L626

    Args:
    record (ndarray): one row of the COSMOS galaxy catalog
    trunc_ratio (float): truncation ratio
    gsparams: an GSParams argument.

    Returns:
    gal: Galsim galaxy
    """

    # record columns:
    # For 'sersicfit', the result is an array of 8 numbers for each:
    #     SERSICFIT[0]: intensity of light profile at the half-light radius.
    #     SERSICFIT[1]: half-light radius measured along the major axis, in
    #                   units of pixels in the COSMOS lensing data reductions
    #                   (0.03 arcsec).
    #     SERSICFIT[2]: Sersic n.
    #     SERSICFIT[3]: q, the ratio of minor axis to major axis length.
    #     SERSICFIT[4]: boxiness, currently fixed to 0, meaning isophotes are
    #                   all elliptical.
    #     SERSICFIT[5]: x0, the central x position in pixels.
    #     SERSICFIT[6]: y0, the central y position in pixels.
    #     SERSICFIT[7]: phi, the position angle in radians. If phi=0, the major
    #                   axis is lined up with the x axis of the image.
    # For 'bulgefit', the result is an array of 16 parameters that comes from
    # doing a 2-component sersic fit.  The first 8 are the parameters for the
    # disk, with n=1, and the last 8 are for the bulge, with n=4.

    bdparams = record["bulgefit"]
    sparams = record["sersicfit"]
    use_bulgefit = record["use_bulgefit"]
    if use_bulgefit:
        # Bulge parameters:
        # Minor-to-major axis ratio:
        bulge_hlr = record["hlr"][1]
        bulge_flux = record["flux"][1]
        disk_hlr = record["hlr"][2]
        disk_flux = record["flux"][2]
        if trunc_ratio <= 0.99:
            btrunc = None
            bulge = galsim.DeVaucouleurs(
                flux=bulge_flux, half_light_radius=bulge_hlr, gsparams=gsparams
            )
            disk = galsim.Exponential(
                flux=disk_flux, half_light_radius=disk_hlr, gsparams=gsparams
            )
        else:
            btrunc = bulge_hlr * trunc_ratio
            bulge = galsim.DeVaucouleurs(
                flux=bulge_flux,
                half_light_radius=bulge_hlr,
                trunc=btrunc,
                gsparams=gsparams,
            )
            dtrunc = disk_hlr * trunc_ratio
            disk = galsim.Sersic(
                1.0,
                flux=disk_flux,
                half_light_radius=disk_hlr,
                trunc=dtrunc,
                gsparams=gsparams,
            )
        # Apply shears for intrinsic shape.
        bulge_q = bdparams[11]
        bulge_beta = bdparams[15] * galsim.radians
        if bulge_q < 1.0:  # pragma: no branch
            bulge = bulge.shear(q=bulge_q, beta=bulge_beta)
        #
        disk_q = bdparams[3]
        disk_beta = bdparams[7] * galsim.radians
        if disk_q < 1.0:  # pragma: no branch
            disk = disk.shear(q=disk_q, beta=disk_beta)
        # Then combine the two components of the galaxy.
        # No center offset is included
        gal = bulge + disk
    else:
        # Do a similar manipulation to the stored quantities for the single
        # Sersic profiles.
        gal_n = sparams[2]
        # Fudge this if it is at the edge of the allowed n values.  Since
        # GalSim (as of #325 and #449) allow Sersic n in the range 0.3<=n<=6,
        # the only problem is that the fits occasionally go as low as n=0.2.
        # The fits in this file only go to n=6, so there is no issue with
        # too-high values, but we also put a guard on that side in case other
        # samples are swapped in that go to higher value of sersic n.
        if gal_n < 0.3:
            gal_n = 0.3
        if gal_n > 6.0:
            gal_n = 6.0

        # GalSim is much more efficient if only a finite number of Sersic n
        # values are used. This (optionally given constructor args) rounds n to
        # the nearest 0.05. change to 0.1 to speed up
        gal_n = _galsim_round_sersic(gal_n, 0.1)
        gal_hlr = record["hlr"][0]
        gal_flux = record["flux"][0]

        if trunc_ratio <= 0.99:
            strunc = None
            gal = galsim.Sersic(
                gal_n,
                flux=gal_flux,
                half_light_radius=gal_hlr,
                gsparams=gsparams,
            )
        else:
            strunc = gal_hlr * trunc_ratio
            gal = galsim.Sersic(
                gal_n,
                flux=gal_flux,
                half_light_radius=gal_hlr,
                trunc=strunc,
                gsparams=gsparams,
            )
        # Apply shears for intrinsic shape.
        gal_q = sparams[3]
        gal_beta = sparams[7] * galsim.radians
        if gal_q < 1.0:
            gal = gal.shear(q=gal_q, beta=gal_beta)
    return gal


def _generate_gal_fft(record, mag_zero, rng, gsparams):
    gal0 = generate_cosmos_gal(record, gsparams=gsparams)
    # E.g., HSC's i-band coadds zero point is 27
    flux = 10 ** ((mag_zero - record["mag_auto"]) / 2.5)
    gal0 = gal0.withFlux(flux)
    # rescale the radius by 'rescale' and keep surface brightness the
    # same
    rescale = rng.np.uniform(0.95, 1.05)
    gal0 = gal0.expand(rescale)
    # rotate by a random angle
    ang = (rng.np.uniform(0.0, np.pi * 2.0)) * galsim.radians
    gal0 = gal0.rotate(ang)
    return gal0


def _generate_gal_mc(record, mag_zero, rng, gsparams, npoints):
    # need to truncate the profile since we do not want
    # Knots locate at infinity
    galp = generate_cosmos_gal_simple(record)
    # accounting for zeropoint difference between COSMOS HST and HSC
    # HSC's i-band coadds zero point is 27
    flux = 10 ** ((mag_zero - record["mag_auto"]) / 2.5)
    galp = galp.withFlux(flux)
    # rescale the radius by 'rescale' and keep surface brightness the
    # same
    rescale = rng.np.uniform(0.95, 1.05)
    galp = galp.expand(rescale)
    # rotate by a random angle
    ang = (rng.np.uniform(0.0, np.pi * 2.0)) * galsim.radians
    galp = galp.rotate(ang)
    gal0 = galsim.RandomKnots(
        npoints=npoints,
        profile=galp,
        rng=rng,
        gsparams=gsparams,
    )
    return gal0


def make_exposure_stamp(
    sim_method,
    rng,
    mag_zero,
    psf_obj,
    scale,
    cat_input,
    ngalx,
    ngaly,
    ngrid,
    rot_field,
    g1,
    g2,
    nrot_per_gal,
    do_shift,
    buff=0,
    draw_method="auto",
):
    ngal = ngalx * ngaly
    gsparams = galsim.GSParams(maximum_fft_size=10240)
    gal0 = None
    gal_image_list = []
    gal_input_list = []
    npix_x = ngalx * ngrid + buff * 2.0
    npix_y = ngaly * ngrid + buff * 2.0
    for _ in range(len(rot_field)):
        gal_image = galsim.ImageF(npix_x, npix_y, scale=scale)
        gal_image.setOrigin(0, 0)
        gal_image_list.append(gal_image)
        gal_input_list.append([])
    for i in range(ngal):
        # boundary
        ix = i % ngalx
        iy = i // ngalx
        # each galaxy
        irot = i % nrot_per_gal
        igal = i // nrot_per_gal
        if irot == 0:
            # prepare the base galaxy
            del gal0
            if sim_method == "fft":
                gal0 = _generate_gal_fft(
                    cat_input[igal],
                    mag_zero,
                    rng,
                    gsparams,
                )
            elif sim_method == "mc":
                gal0 = _generate_gal_mc(
                    cat_input[igal],
                    mag_zero,
                    rng,
                    gsparams,
                    npoints=30,
                )
            else:
                raise ValueError("Does not support sim_method=%s" % sim_method)
        else:
            # rotate the base galaxy
            assert gal0 is not None
            ang = np.pi / nrot_per_gal * galsim.radians
            # update gal0
            gal0 = gal0.rotate(ang)

        s01 = rng.np.uniform(low=-0.5, high=0.5) * scale
        s02 = rng.np.uniform(low=-0.5, high=0.5) * scale
        for ii, rr in enumerate(rot_field):
            # base rotation and shear distortion
            gal = gal0.rotate(rr * galsim.radians).shear(g1=g1, g2=g2)
            gal = galsim.Convolve([psf_obj, gal], gsparams=gsparams)
            b = galsim.BoundsI(
                ix * ngrid + buff,
                (ix + 1) * ngrid - 1 + buff,
                iy * ngrid + buff,
                (iy + 1) * ngrid - 1 + buff,
            )
            sub_img = gal_image_list[ii][b]
            # shift with a random offset
            s1, s2 = coord_rotate(s01, s02, xref=0, yref=0, theta=rr)
            if do_shift:
                gal = gal.shift(s1, s2)
            # shift to (ngrid//2, ngrid//2)
            # since I set it as the default center of grids in the simulation
            gal = gal.shift(0.5 * scale, 0.5 * scale)
            gal.drawImage(sub_img, add_to_image=True, method=draw_method)
    gc.collect()
    exposures = [img.array for img in gal_image_list]
    return exposures


def read_cosmos_catalog(filename):
    # read the input COSMOS galaxy catalog
    cat_input = fitsio.read(filename)
    return cat_input


class CosmosCatalog(object):
    def __init__(
        self,
        filename=None,
        max_mag=None,
        min_mag=None,
        max_hlr=None,
        min_hlr=None,
        gal_type="mixed",
        logger=None,
    ):
        if logger is None:
            logger = setup_custom_logger(verbose=False)
        src = read_cosmos_catalog(filename)
        # Initializing selector mask with all Trues
        sel = np.ones(len(src), dtype=bool)
        # Filtering conditions
        if max_mag is not None:
            sel &= src["mag_auto"] < max_mag
        if min_mag is not None:
            sel &= src["mag_auto"] >= min_mag
        if max_hlr is not None:
            if gal_type == "mixed":
                sel &= (src["hlr"][:, 0:3] < max_hlr).all(axis=1)
            elif gal_type == "sersic":
                sel &= src["hlr"][:, 0] < max_hlr
            elif gal_type == "bulgedisk":
                sel &= (src["hlr"][:, 1:3] < max_hlr).all(axis=1)
        if min_hlr is not None:
            if gal_type == "mixed":
                sel &= (src["hlr"][:, 0:3] >= max_hlr).all(axis=1)
            elif gal_type == "sersic":
                sel &= src["hlr"][:, 0] >= min_hlr
            elif gal_type == "bulgedisk":
                sel &= (src["hlr"][:, 1:3] >= max_hlr).all(axis=1)
        # Applying selector mask
        src = src[sel]

        if gal_type == "mixed":
            logger.info("Creating Mixed galaxy profiles")
        elif gal_type == "sersic":
            logger.info("Creating single Sersic profiles")
            src = src[src["use_bulgefit"] == 0]
        elif gal_type == "bulgedisk":
            logger.info("Creating Bulge + Disk profiles")
            src = src[src["use_bulgefit"] != 0]
        elif gal_type == "debug":
            # This is used for debug
            logger.info("Creating profile for debug")
            src = src[src["use_bulgefit"] == 0]
            # src["sersicfit"][:, 3] = 1.0  # round galaxies
            # src["sersicfit"][:, 2] = 0.5  # only Gaussian
            src["sersicfit"][:, 2] = 1.0  # only Exponential
        else:
            raise ValueError("Do not support gal_type=%s" % gal_type)
        self.cat_input = src
        self.ntrain = len(src)
        return

    def make_catalog(self, rng, n):
        if self.ntrain < n:
            raise ValueError(
                "There is not enough number of galaxies",
            )
        # nrot_per_gal is the number of rotated galaxies in each subfield
        inds = rng.np.integers(low=0, high=self.ntrain, size=n)
        src = self.cat_input[inds]
        return src


def make_isolated_sim(
    ny,
    nx,
    psf_obj,
    gname,
    seed,
    cat_name=None,
    scale=0.168,
    mag_zero=27.0,
    rot_field=None,
    shear_value=0.02,
    ngrid=64,
    nrot_per_gal=4,
    max_mag=None,
    min_mag=None,
    max_hlr=None,
    min_hlr=None,
    gal_type="mixed",
    do_shift=False,
    npoints=30,
    sim_method="fft",
    buff=0,
    draw_method="auto",
    return_catalog=False,
    verbose=False,
):
    """Makes basic **isolated** galaxy image simulation.

    Args:
    ny (int): number of pixels in y direction
    nx (int): number of pixels in y direction
    psf_obj (PSF): input PSF object of galsim
    gname (str): shear distortion setup
    seed (int): index of the simulation
    cat_name (str): input catalog name
    scale (float): pixel scale
    mag_zero (float): magnitude zero point [27 for HSC]
    rot_field (list): additional rotation angle
    shear_value (float): shear distortion amplitude
    ngrid (int): stampe size
    nrot_per_gal (int): number of rotations
    max_mag (float): maximum magnitude cut
    min_mag (float): minimum magnitude cut
    max_hlr (float): maximum half light radius cut [arcsec]
    min_hlr (float): minimum half light radius cut [arcsec]
    gal_type (float): galaxy morphology (mixed, sersic, or bulgedisk)
    do_shift (bool): whether do shfits
    npoints (int): number of random points when
    sim_method (str): galaxy tpye ("fft" or "mc")
    buff (int): buff size (zero padding near boundaries)
    draw_method (str): method used for drawing image by Galsim
    return_catalog (bool): whether returning the input catalog
    verbose (bool): whether show log info
    """

    logger = setup_custom_logger(verbose=verbose)
    if nx % ngrid != 0:
        raise ValueError("nx is not divisible by ngrid")
    if ny % ngrid != 0:
        raise ValueError("ny is not divisible by ngrid")
    # Basic parameters
    ngalx = int(nx // ngrid)
    ngaly = int(ny // ngrid)
    ngal = ngalx * ngaly
    rng = galsim.BaseDeviate(seed)
    ngeff = max(ngal // nrot_per_gal, 1)
    if cat_name is None:
        cat_name = os.path.join(_data_dir, "src_cosmos.fits")
    cosmos_cat = CosmosCatalog(
        filename=cat_name,
        max_mag=max_mag,
        min_mag=min_mag,
        max_hlr=max_hlr,
        min_hlr=min_hlr,
        gal_type=gal_type,
        logger=logger,
    )
    cat_input = cosmos_cat.make_catalog(rng=rng, n=ngeff)
    # Get the shear information
    shear_list = np.array([-shear_value, shear_value, 0.0])
    dis_version = int(eval(gname.split("-")[-1]))
    assert dis_version < 3 and dis_version >= 0, "gname is not supported"
    shear_const = shear_list[dis_version]
    g_version = gname.split("-")[0]
    if g_version == "g1":
        g1 = shear_const
        g2 = 0.0
    elif g_version == "g2":
        g1 = 0.0
        g2 = shear_const
    elif g_version == "g1_g2":
        g1 = shear_const
        g2 = 0.02
    elif g_version == "g2_g1":
        g1 = 0.02
        g2 = shear_const
    else:
        raise ValueError("cannot decide g1 or g2")
    logger.info("Processing for %s, and shear is %s." % (gname, shear_const))

    if rot_field is None:
        rot_field = [0.0]
    exposures = make_exposure_stamp(
        sim_method=sim_method,
        rng=rng,
        mag_zero=mag_zero,
        psf_obj=psf_obj,
        scale=scale,
        cat_input=cat_input,
        ngalx=ngalx,
        ngaly=ngaly,
        ngrid=ngrid,
        rot_field=rot_field,
        g1=g1,
        g2=g2,
        nrot_per_gal=nrot_per_gal,
        do_shift=do_shift,
        buff=buff,
        draw_method=draw_method,
    )
    if not return_catalog:
        return exposures
    else:
        return exposures, cat_input


def make_noise_sim(
    out_dir,
    infname,
    ind0,
    ny=6400,
    nx=6400,
    scale=0.168,
    verbose=False,
):
    """Makes pure noise for galaxy image simulation.

    Args:
    out_dir (str): output directory
    ind0 (int): index of the simulation
    ny (int): number of pixels in y direction
    nx (int): number of pixels in x direction
    scale (float): pixel scale
    verbose (bool): whether show log info
    """
    logger = setup_custom_logger(verbose=verbose)
    logger.info("begining for field %04d" % (ind0))
    logger.info("simulating noise for field %s" % (ind0))
    variance = 0.01
    ud = galsim.UniformDeviate(ind0 * 10000 + 1)

    # setup the galaxy image and the noise image
    noi_image = galsim.ImageF(nx, ny, scale=scale)
    noi_image.setOrigin(0, 0)
    noise_obj = galsim.getCOSMOSNoise(
        file_name=infname, rng=ud, cosmos_scale=scale, variance=variance
    )
    noise_obj.applyTo(noi_image)
    return noi_image.array


def make_blended_sim(
    out_dir,
    psf_obj,
    gname,
    ind0,
    cat_name=None,
    ny=5000,
    nx=5000,
    rfrac=0.46,
    scale=0.168,
    mag_zero=27.0,
    rot_field=0.0,
    shear_value=0.02,
    nrot=4,
    rescale_min_max=None,
    verbose=False,
):
    """Makes cosmo-like blended galaxy image simulations.

    Args:
    out_dir (str): output directory
    psf_obj (PSF): input PSF object of galsim
    gname (str): shear distortion setup
    ind0 (int): index of the simulation
    cat_name (str): input catalog Name [default: COSMOS 25.2 catalog]
    ny (int): number of galaxies in y direction [default: 5000]
    nx (int): number of galaxies in x direction [default: 5000]
    rfrac(float): fraction of radius to minimum between nx and ny
    mag_zero (float): magnitude zero point
    rot_field (float): additional rotational angle [in units of radians]
    shear_value (float): amplitude of the input shear
    nrot (int): number of rotation, optional
    rescale_min_max (list|ndarray): lower and upper bounds of galaxy size
        rescaling factor, optional
    verbose (bool): whether show log info
    """

    logger = setup_custom_logger(verbose=verbose)
    np.random.seed(ind0)

    bigfft = galsim.GSParams(maximum_fft_size=10240)  # galsim setup
    # Get the shear information
    # Three choice on g(-shear_value,0,shear_value)
    shear_list = np.array([-shear_value, 0.0, shear_value])
    shear_list = shear_list[[eval(i) for i in gname.split("-")[-1]]]

    # number of galaxy
    # we only have `ngeff' galsim galaxies but with `nrot' rotation
    r2 = (min(nx, ny) * rfrac) ** 2.0
    density = int(out_dir.split("_psf")[0].split("_cosmo")[-1])
    ngal = max(int(r2 * np.pi * scale**2.0 / 3600.0 * density), nrot)
    ngal = int(ngal // (nrot * 2) * (nrot * 2))
    ngeff = ngal // (nrot * 2)
    logger.info(
        "We have %d galaxies in total, and each %d are the same" % (ngal, nrot)
    )
    if cat_name is None:
        cat_name = os.path.join(_data_dir, "src_cosmos.fits")

    # get the cosmos catalog
    cat_input = fitsio.read(cat_name)
    ntrain = len(cat_input)
    inds = np.random.randint(0, ntrain, ngeff)
    cat_input = cat_input[inds]

    # evenly distributed within a radius, min(nx,ny)*rfrac
    rarray = np.sqrt(r2 * np.random.rand(ngeff))  # radius
    tarray = np.random.uniform(0.0, np.pi / nrot, ngeff)  # theta (0,pi/nrot)
    tarray = tarray + rot_field
    xarray = rarray * np.cos(tarray) + nx // 2  # x
    yarray = rarray * np.sin(tarray) + ny // 2  # y
    if rescale_min_max is None:
        rescale_min_max = [0.95, 1.05]
    rsarray = np.random.uniform(rescale_min_max[0], rescale_min_max[1], ngeff)
    del rarray, tarray

    zbound = np.array([1e-5, 0.5477, 0.8874, 1.3119, 12.0])  # sim 3

    gal_image = galsim.ImageF(nx, ny, scale=scale)
    gal_image.setOrigin(0, 0)
    for ii in range(ngal):
        ig = ii // (nrot * 2)
        irot = ii % (nrot * 2)
        ss = cat_input[ig]
        # x,y
        xi = xarray[ig]
        yi = yarray[ig]
        # randomly rotate by an angle; we have 180/nrot deg pairs to remove
        # shape noise in additive bias estimation
        rot_ang = np.pi / nrot * irot
        ang = rot_ang * galsim.radians
        xi, yi = coord_rotate(xi, yi, nx // 2, ny // 2, rot_ang)
        # determine redshift
        shear_inds = np.where(
            (ss["zphot"] > zbound[:-1]) & (ss["zphot"] <= zbound[1:])
        )[0]
        if len(shear_inds) == 1:
            if gname.split("-")[0] == "g1":
                g1 = shear_list[shear_inds][0]
                g2 = 0.0
            elif gname.split("-")[0] == "g2":
                g1 = 0.0
                g2 = shear_list[shear_inds][0]
            else:
                raise ValueError("g1 or g2 must be in gname")
        else:
            g1 = 0.0
            g2 = 0.0

        gal = generate_cosmos_gal(ss, gsparams=bigfft)
        # rescale the radius while keeping the surface brightness the same
        gal = gal.expand(rsarray[ig])
        # determine and assign flux
        # (HSC's i-band coadds zero point is 27)
        # (LSST's coadds zero point is 30)
        flux = 10 ** ((mag_zero - ss["mag_auto"]) / 2.5)
        gal = gal.withFlux(flux)
        # rotate by 'ang'
        gal = gal.rotate(ang)
        # lensing shear
        gal = gal.shear(g1=g1, g2=g2)
        # position and subpixel offset
        xi, yi = coord_distort_1(xi, yi, nx // 2, ny // 2, g1, g2)
        xu = int(xi)
        yu = int(yi)
        dx = (0.5 + xi - xu) * scale
        dy = (0.5 + yi - yu) * scale
        gal = gal.shift(dx, dy)
        # PSF
        gal = galsim.Convolve([psf_obj, gal], gsparams=bigfft)
        # Bounary
        r_grid = max(gal.getGoodImageSize(scale), 32)
        rx1 = np.min([r_grid, xu])
        rx2 = np.min([r_grid, nx - xu - 1])
        rx = int(min(rx1, rx2))
        del rx1, rx2
        ry1 = np.min([r_grid, yu])
        ry2 = np.min([r_grid, ny - yu - 1])
        ry = int(min(ry1, ry2))
        del ry1, ry2
        # draw galaxy
        b = galsim.BoundsI(xu - rx, xu + rx - 1, yu - ry, yu + ry - 1)
        sub_img = gal_image[b]
        gal.drawImage(sub_img, add_to_image=True)
        del gal, b, sub_img, xu, yu, xi, yi, r_grid
    gc.collect()
    del cat_input, psf_obj
    return gal_image.array



from typing import Any

import numpy as np
from numpy.typing import NDArray


def resize_array(array: NDArray[Any], target_shape: tuple[int, int] = (64, 64)):
    """This is a util function to resize array to the target shape
    Args:
    array (NDArray): input array
    target_shape (tuple): output array's shape

    Returns:
    array (NDArray): output array with the target shape
    """
    target_height, target_width = target_shape
    input_height, input_width = array.shape

    # Crop if larger
    if input_height > target_height:
        start_h = (input_height - target_height) // 2
        array = array[start_h : start_h + target_height, :]
    if input_width > target_width:
        start_w = (input_width - target_width) // 2
        array = array[:, start_w : start_w + target_width]

    # Pad with zeros if smaller
    if input_height < target_height:
        pad_height = target_height - input_height
        pad_top = pad_height // 2
        pad_bottom = pad_height - pad_top
        array = np.pad(
            array,
            ((pad_bottom, pad_top), (0, 0)),
            mode="constant",
            constant_values=0.0,
        )

    if input_width < target_width:
        pad_width = target_width - input_width
        pad_right = pad_width // 2
        pad_left = pad_width - pad_right
        array = np.pad(
            array,
            ((0, 0), (pad_left, pad_right)),
            mode="constant",
        )
    return array


def rotate90(image):
    rotated_image = np.zeros_like(image)
    rotated_image[1:, 1:] = np.rot90(m=image[1:, 1:], k=-1)
    return rotated_image



# Image Tasks of Shapelet Based Measurements
#
# python lib
import numpy as np
import numpy.lib.recfunctions as rfn
from numpy.typing import NDArray
from pydantic import BaseModel, Field

from . import BasePsf, FpfsImage, mask_galaxy_image, measure_fpfs
from .base import FpfsKernel

npix_patch = 256
npix_overlap = 64
npix_default = 64


class FpfsTask(FpfsKernel):
    """A base class for measurement

    Args:
    npix (int): number of pixels in a postage stamp
    pixel_scale (float): pixel scale in arcsec
    sigma_arcsec (float): Shapelet kernel size
    noise_variance (float): variance of image noise
    kmax (float | None): maximum k
    psf_array (ndarray): an average PSF image [default: None]
    kmax_thres (float): the tuncation threshold on Gaussian [default: 1e-20]
    do_detection (bool): whether compute detection kernels
    bound (int): minimum distance to boundary [default: 0]
    verbose (bool): whether display INFO
    """

    def __init__(
        self,
        *,
        npix: int,
        pixel_scale: float,
        sigma_arcsec: float,
        noise_variance: float = -1,
        kmax: float | None = None,
        psf_array: NDArray | None = None,
        kmax_thres: float = 1e-20,
        do_detection: bool = True,
        bound: int = 0,
        verbose: bool = False,
    ) -> None:
        super().__init__(
            npix=npix,
            pixel_scale=pixel_scale,
            sigma_arcsec=sigma_arcsec,
            kmax=kmax,
            psf_array=psf_array,
            kmax_thres=kmax_thres,
            do_detection=do_detection,
            verbose=verbose,
        )
        self.prepare_fpfs_bases()
        self.mtask = FpfsImage(
            nx=self.npix,
            ny=self.npix,
            scale=self.pixel_scale,
            sigma_arcsec=self.sigma_arcsec,
            klim=self.kmax / self.pixel_scale,
            psf_array=self.psf_array,
            use_estimate=True,
        )
        # print(self.kmax / self.pixel_scale)
        # print(self.kmax)

        if self.do_detection:
            self.dtask = FpfsImage(
                nx=npix_patch,
                ny=npix_patch,
                scale=self.pixel_scale,
                sigma_arcsec=self.sigma_arcsec,
                klim=self.kmax / self.pixel_scale,
                psf_array=self.psf_array,
                use_estimate=True,
                npix_overlap=npix_overlap,
                bound=bound,
            )
            if not noise_variance > 0:
                raise ValueError("Noise variance should be positive")
            self.prepare_covariance(variance=noise_variance)
        else:
            self.dtask = None

        return

    def detect(
        self,
        *,
        gal_array: NDArray,
        fthres: float,
        pthres: float,
        omega_v: float,
        v_min: float,
        noise_array: NDArray | None = None,
        mask_array: NDArray | None = None,
        star_cat: NDArray | None = None,
    ) -> NDArray:
        """This function detects galaxy from image

        Args:
        gal_array (NDArray): galaxy image data
        fthres (float): flux threshold
        pthres (float): peak threshold
        omega_v (float): smoothness parameter for pixel difference
        noise_array (NDArray|None): pure noise image
        mask_array (NDArray|None): mask image
        star_cat (NDArray|None): bright star catalog

        Returns:
        (NDArray): galaxy detection catalog
        """
        if mask_array is not None:
            # Set the value inside star mask to zero
            mask_galaxy_image(gal_array, mask_array, True, star_cat)
            if noise_array is not None:
                # Also do it for pure noise image
                mask_galaxy_image(noise_array, mask_array, True, star_cat)
        assert self.dtask is not None
        return self.dtask.detect_source(
            gal_array=gal_array,
            fthres=fthres,
            pthres=pthres,
            std_m00=self.std_m00 * self.pixel_scale**2.0,
            omega_v=omega_v * self.pixel_scale**2.0,
            v_min=v_min * self.pixel_scale**2.0,
            noise_array=noise_array,
            mask_array=mask_array,
        )

    def run_psf_array(
        self,
        *,
        gal_array: NDArray,
        psf_array: NDArray,
        det: NDArray | None = None,
        noise_array: NDArray | None = None,
    ) -> tuple[NDArray, NDArray | None]:
        """This function measure galaxy shapes at the position of the detection
        using PSF image data

        Args:
        gal_array (NDArray): galaxy image data
        psf_array (NDArray): psf image data
        det (list|None): detection catalog
        noise_array (NDArray | None): noise image data [default: None]

        Returns:
        src_g (NDArray): source measurement catalog
        src_n (NDArray): noise measurement catalog
        """
        # self.logger.warning("Input PSF is array")
        src_g = self.mtask.measure_source(
            gal_array=gal_array,
            filter_image=self.bfunc_use,
            psf_array=psf_array,
            det=det,
            do_rotate=False,
        )
        if noise_array is not None:
            src_n = self.mtask.measure_source(
                gal_array=noise_array,
                filter_image=self.bfunc_use,
                psf_array=psf_array,
                det=det,
                do_rotate=True,
            )
            src_g = src_g + src_n
        else:
            src_n = None
        return src_g, src_n

    def run_psf_python(
        self,
        gal_array: NDArray,
        psf_obj: BasePsf,
        det: NDArray,
        noise_array: NDArray | None = None,
    ) -> tuple[NDArray, NDArray | None]:
        """This function measure galaxy shapes at the position of the detection
        using PSF image data

        Args:
        gal_array (NDArray): galaxy image data
        psf_obj (BasePsf): PSF object in python
        noise_array (NDArray | None): noise image data [default: None]
        det (list|None): detection catalog

        Returns:
        src_g (NDArray): source measurement catalog
        src_n (NDArray): noise measurement catalog
        """
        # self.logger.warning("Input PSF is python object")
        det_dtype = det.dtype
        src_g = []
        src_n = []
        for _d in det:
            this_psf_array = psf_obj.draw(x=_d["x"], y=_d["y"])
            # TODO: remove det_array
            det_array = np.array([_d], dtype=det_dtype)
            srow = self.mtask.measure_source(
                gal_array=gal_array,
                filter_image=self.bfunc_use,
                psf_array=this_psf_array,
                det=det_array,
                do_rotate=False,
            )[0]
            if noise_array is not None:
                nrow = self.mtask.measure_source(
                    gal_array=noise_array,
                    filter_image=self.bfunc_use,
                    psf_array=this_psf_array,
                    det=det_array,
                    do_rotate=True,
                )[0]
                srow = srow + nrow
                src_n.append(nrow)
            src_g.append(srow)
        if len(src_n) == 0:
            src_n = None
        else:
            assert len(src_n) == len(src_g)
            src_n = np.array(src_n)
        src_g = np.array(src_g)
        return src_g, src_n

    def run_psf_cpp(
        self,
        gal_array: NDArray,
        psf_obj: BasePsf,
        noise_array: NDArray | None = None,
        det: NDArray | None = None,
    ) -> tuple[NDArray, NDArray | None]:
        """This function measure galaxy shapes at the position of the detection
        using PSF model with spacial variation

        Args:
        gal_array (NDArray): galaxy image data
        psf_obj (BasePsf): psf object in cpp
        noise_array (NDArray | None): noise image data [default: None]
        det (list|None): detection catalog

        Returns:
        src_g (NDArray): source measurement catalog
        src_n (NDArray): noise measurement catalog
        """

        # self.logger.warning("Input PSF is cpp object")
        src_g = self.mtask.measure_source(
            gal_array=gal_array,
            filter_image=self.bfunc_use,
            psf_obj=psf_obj,
            det=det,
            do_rotate=False,
        )
        if noise_array is not None:
            src_n = self.mtask.measure_source(
                gal_array=noise_array,
                filter_image=self.bfunc_use,
                psf_obj=psf_obj,
                det=det,
                do_rotate=True,
            )
            src_g = src_g + src_n
        else:
            src_n = None
        return src_g, src_n

    def run(
        self,
        gal_array: NDArray,
        psf: BasePsf | NDArray,
        det: NDArray | None = None,
        noise_array: NDArray | None = None,
    ):
        """This function measure galaxy shapes at the position of the detection

        Args:
        gal_array (NDArray): galaxy image data
        det (NDArray): detection catalog
        psf (BasePsf | NDArray): psf image data or psf model
        noise_array (NDArray | None): noise image data [default: None]

        Returns:
        (NDArray): galaxy measurement catalog
        """
        if isinstance(psf, np.ndarray):
            src_g, src_n = self.run_psf_array(
                gal_array=gal_array,
                psf_array=psf,
                noise_array=noise_array,
                det=det,
            )
        elif isinstance(psf, BasePsf):
            assert det is not None
            if psf.crun:
                # For the case PSF is a C++ object
                src_g, src_n = self.run_psf_cpp(
                    gal_array=gal_array,
                    psf_obj=psf,
                    noise_array=noise_array,
                    det=det,
                )
            else:
                # For the case PSF is a Python object
                src_g, src_n = self.run_psf_python(
                    gal_array=gal_array,
                    psf_obj=psf,
                    noise_array=noise_array,
                    det=det,
                )

        else:
            raise RuntimeError("psf does not have a correct type")
        src_g = rfn.unstructured_to_structured(
            arr=src_g,
            dtype=self.dtype,
        )
        if src_n is not None:
            src_n = rfn.unstructured_to_structured(arr=src_n, dtype=self.dtype)

        return {
            "data": src_g,
            "noise": src_n,
        }


class FpfsConfig(BaseModel):
    npix: int = Field(
        default=64,
        description="""size of the stamp before Fourier Transform
        """,
    )
    kmax_thres: float = Field(
        default=1e-12,
        description="""The threshold used to define the upper limit of k we use
        in Fourier space.
        """,
    )
    bound: int = Field(
        default=35,
        description="""Boundary buffer length, the sources in the buffer reion
        are not counted.
        """,
    )
    sigma_arcsec: float = Field(
        default=0.52,
        description="""Smoothing scale of the shapelet and detection kernel.
        """,
    )
    sigma_arcsec1: float = Field(
        default=-1,
        description="""Smoothing scale of the second shapelet kernel.
        """,
    )
    sigma_arcsec2: float = Field(
        default=-1,
        description="""Smoothing scale of the third shapelet kernel.
        """,
    )
    pthres: float = Field(
        default=0.12,
        description="""Detection threshold (peak identification) for the
        pooling.
        """,
    )
    fthres: float = Field(
        default=8.0,
        description="""Detection threshold (minimum signal-to-noise ratio) for
        the first pooling.
        """,
    )
    omega_r2: float = Field(
        default=4.8,
        description="""
        smoothness parameter for r2 cut
        """,
    )
    r2_min: float = Field(
        default=0.1,
        description="""Minimum trace moment matrix
        """,
    )
    omega_v: float = Field(
        default=0.9,
        description="""
        smoothness parameter for v cut
        """,
    )
    v_min: float = Field(
        default=0.45,
        description="""Minimum of v
        """,
    )
    snr_min: float = Field(
        default=12,
        description="""Minimum Signal-to-Noise Ratio for detection.
        """,
    )
    c0: float = Field(
        default=8.4,
        description="""Weighting parameter for m00 for ellipticity definition.
        """,
    )


def process_image(
    *,
    fpfs_config: FpfsConfig,
    pixel_scale: float,
    noise_variance: float,
    mag_zero: float,
    gal_array: NDArray,
    psf_array: NDArray,
    noise_array: NDArray | None = None,
    mask_array: NDArray | None = None,
    star_catalog: NDArray | None = None,
    detection: NDArray | None = None,
    psf_object: BasePsf | None = None,
    do_compute_detect_weight: bool = True,
    only_return_detection_modes: bool = False,
    base_column_name: str | None = None,
):
    """Run measurement algorithms on the input exposure, and optionally
    populate the resulting catalog with extra information.

    Args:
    fpfs_config (FpfsConfig):  configuration object
    pixel_scale (float): pixel scale in arcsec
    noise_variance (float): variance of image noise
    mag_zero (float): magnitude zero point
    do_detection (bool): whether compute detection kernel
    gal_array (NDArray[float64]): galaxy exposure array
    psf_array (ndarray): an average PSF image
    noise_array (NDArray | None): pure noise array
    mask_array (NDArray | None): mask array (1 for masked)
    star_catalog (NDArray | None): bright star catalog
    detection (NDArray | None): detection catalog
    psf_object (BasePsf | None): PSF object
    do_compute_detect_weight (bool): whether to compute detection weight
    only_return_detection_modes (bool): only return linear modes for detection
    base_column_name (str | None): base column name

    Returns:
    (NDArray) FPFS catalog
    """
    if only_return_detection_modes:
        assert do_compute_detect_weight

    ratio = 1.0 / (10 ** ((30 - mag_zero) / 2.5))
    r2_min = fpfs_config.r2_min * ratio
    omega_r2 = fpfs_config.omega_r2 * ratio
    v_min = fpfs_config.v_min * ratio
    omega_v = fpfs_config.omega_v * ratio
    fpfs_c0 = fpfs_config.c0 * ratio

    if psf_object is None:
        psf_object = psf_array

    out_list = []

    if do_compute_detect_weight or (detection is None):
        ftask = FpfsTask(
            npix=fpfs_config.npix,
            pixel_scale=pixel_scale,
            sigma_arcsec=fpfs_config.sigma_arcsec,
            noise_variance=noise_variance,
            psf_array=psf_array,
            kmax_thres=fpfs_config.kmax_thres,
            do_detection=True,
            bound=fpfs_config.bound,
        )
        std_m00 = ftask.std_m00
        m00_min = fpfs_config.snr_min * std_m00
        if detection is None:
            detection = ftask.detect(
                gal_array=gal_array,
                fthres=fpfs_config.fthres,
                pthres=fpfs_config.pthres,
                noise_array=noise_array,
                v_min=v_min,
                omega_v=omega_v,
                mask_array=mask_array,
                star_cat=star_catalog,
            )
        else:
            det_names = ("y", "x", "is_peak", "mask_value")
            if detection.dtype.names != det_names:
                raise ValueError("detection has wrong cloumn names")
        out_list.append(detection)

        if do_compute_detect_weight:
            src = ftask.run(
                gal_array=gal_array,
                psf=psf_object,
                det=detection,
                noise_array=noise_array,
            )
            if only_return_detection_modes:
                return src
            meas = measure_fpfs(
                C0=fpfs_c0,
                v_min=v_min,
                omega_v=omega_v,
                pthres=fpfs_config.pthres,
                m00_min=m00_min,
                std_m00=std_m00,
                r2_min=r2_min,
                omega_r2=omega_r2,
                x_array=src["data"],
                y_array=src["noise"],
            )
            del src
            map_dict = {name: "fpfs_" + name for name in meas.dtype.names}
            out_list.append(rfn.rename_fields(meas, map_dict))

        del ftask

    if fpfs_config.sigma_arcsec1 > 0:
        ftask = FpfsTask(
            npix=fpfs_config.npix,
            pixel_scale=pixel_scale,
            sigma_arcsec=fpfs_config.sigma_arcsec1,
            psf_array=psf_array,
            kmax_thres=fpfs_config.kmax_thres,
            do_detection=False,
        )
        src = ftask.run(
            gal_array=gal_array,
            psf=psf_object,
            det=detection,
            noise_array=noise_array,
        )
        meas1 = measure_fpfs(
            C0=fpfs_c0,
            x_array=src["data"],
            y_array=src["noise"],
        )
        del src, ftask
        map_dict = {name: "fpfs1_" + name for name in meas1.dtype.names}
        out_list.append(rfn.rename_fields(meas1, map_dict))
        del meas1

    if fpfs_config.sigma_arcsec2 > 0:
        ftask = FpfsTask(
            npix=fpfs_config.npix,
            pixel_scale=pixel_scale,
            sigma_arcsec=fpfs_config.sigma_arcsec2,
            psf_array=psf_array,
            kmax_thres=fpfs_config.kmax_thres,
            do_detection=False,
        )
        src = ftask.run(
            gal_array=gal_array,
            psf=psf_object,
            det=detection,
            noise_array=noise_array,
        )
        meas2 = measure_fpfs(
            C0=fpfs_c0,
            x_array=src["data"],
            y_array=src["noise"],
        )
        del src, ftask
        map_dict = {name: "fpfs2_" + name for name in meas2.dtype.names}
        out_list.append(rfn.rename_fields(meas2, map_dict))
        del meas2

    result = rfn.merge_arrays(
        out_list,
        flatten=True,
        usemask=False,
    )
    if base_column_name is not None:
        assert result.dtype.names is not None
        map_dict = {
            name: base_column_name + name for name in result.dtype.names
        }
        result = rfn.rename_fields(result, map_dict)

    return result



import ctypes
import os
import sys

import pybind11
from setuptools import Extension, find_packages, setup

conda_prefix = os.environ.get("CONDA_PREFIX")
include_dirs = ["include/"]
if conda_prefix:
    include_dirs.append(os.path.join(conda_prefix, "include"))


include_dirs.append(pybind11.get_include())


def find_fftw_lib():
    import distutils.sysconfig

    try_libdirs = []

    # Start with the explicit FFTW_DIR, if present.
    if "FFTW_DIR" in os.environ:
        try_libdirs.append(os.environ["FFTW_DIR"])
        try_libdirs.append(os.path.join(os.environ["FFTW_DIR"], "lib"))

    # Add the python system library directory.
    try_libdirs.append(distutils.sysconfig.get_config_var("LIBDIR"))

    # Try some standard locations where things get installed
    try_libdirs.extend(["/usr/local/lib", "/usr/lib"])
    if sys.platform == "darwin":
        try_libdirs.extend(["/sw/lib", "/opt/local/lib"])

    # Check the directories in LD_LIBRARY_PATH.
    # This doesn't work on OSX >= 10.11
    for path in ["LIBRARY_PATH", "LD_LIBRARY_PATH", "DYLD_LIBRARY_PATH"]:
        if path in os.environ:
            for dir in os.environ[path].split(":"):
                try_libdirs.append(dir)

    # The user's home directory is often a good place to check.
    try_libdirs.append(os.path.join(os.path.expanduser("~"), "lib"))
    if sys.platform == "darwin":
        lib_ext = ".dylib"
    else:
        lib_ext = ".so"
    name = "libfftw3" + lib_ext
    tried_dirs = set()  # Keep track, so we don't try the same thing twice.
    for dir in try_libdirs:
        if dir == "":
            continue  # This messes things up if it's in there.
        if dir in tried_dirs:
            continue
        else:
            tried_dirs.add(dir)
        if not os.path.isdir(dir):
            continue
        libpath = os.path.join(dir, name)
        if not os.path.isfile(libpath):
            continue
        try:
            lib = ctypes.cdll.LoadLibrary(libpath)
            return libpath
        except OSError as e:
            # Some places use lib64 rather than/in addition to lib.
            if dir.endswith("lib") and os.path.isdir(dir + "64"):
                dir += "64"
                try:
                    libpath = os.path.join(dir, name)
                    if not os.path.isfile(libpath):
                        continue
                    lib = ctypes.cdll.LoadLibrary(libpath)
                    return libpath
                except OSError:
                    pass

    # If we didn't find it anywhere, but the user has set FFTW_DIR, trust it.
    if "FFTW_DIR" in os.environ:
        libpath = os.path.join(os.environ["FFTW_DIR"], name)
        print("WARNING:")
        print("Could not find an installed fftw3 library named %s" % name)
        print(
            "Trust the provided FFTW_DIR=%s for the library location." % libpath
        )
        print("If this is incorrect, you may have errors later when linking.")
        return libpath


fftw_lib = find_fftw_lib()
if fftw_lib is not None:
    fftw_libpath, fftw_libname = os.path.split(fftw_lib)
    include_dirs.append(
        os.path.join(
            os.path.split(fftw_libpath)[0],
            "include",
        )
    )

ext_modules = []
ext_modules.append(
    Extension(
        "anacal._anacal",  # Name of the module
        [
            "python/anacal/_anacalLib.cc",
            "src/image.cpp",
            "src/model.cpp",
            "src/fpfs.cpp",
            "src/fpfs/image.cpp",
            "src/fpfs/catalog.cpp",
            "src/noise.cpp",
            "src/psf.cpp",
            "src/mask.cpp",
            "src/math.cpp",
            "src/ngmix.cpp",
            "src/table.cpp",
            "src/detector.cpp",
            "src/geometry.cpp",
            "src/task.cpp",
        ],
        include_dirs=include_dirs,
        libraries=["fftw3"],
        language="c++",
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Wdeprecated-declarations",
            "-std=c++17",
            "-fopenmp",
            "-O3",
            "-fvisibility=hidden",
        ],
        extra_link_args=["-flto", "-fopenmp"],
    )
)

this_dir = os.path.dirname(os.path.realpath(__file__))
__version__ = ""
fname = os.path.join(this_dir, "python/anacal", "__version__.py")
with open(fname, "r") as ff:
    exec(ff.read())
long_description = open(os.path.join(this_dir, "README.md")).read()

setup(
    name="anacal",
    version=__version__,
    author="Xiangchong Li",
    author_email="mr.superonion@hotmail.com",
    python_requires=">=3.10",
    setup_requires=["pybind11>=2.2", "setuptools>=38", "wheel"],
    install_requires=[
        "pybind11>=2.2",
        "numpy",
        "galsim",
        "fitsio",
        "pydantic",
    ],
    packages=find_packages(where="python"),
    package_dir={"": "python"},
    include_package_data=True,
    zip_safe=False,
    package_data={
        "anacal": ["data/*.fits"],
    },
    ext_modules=ext_modules,
    url="https://github.com/mr-superonion/AnaCal/",
    long_description=long_description,
    long_description_content_type="text/markdown",
)
